# Detect suitable CLANG library
unless ENV['LIBCLANG']
    begin
        libdir = %x{llvm-config --libdir}.chomp
        libs = Dir.glob( "#{libdir}/*libclang*so*" )
        ENV['LIBCLANG'] = libs[0]
    rescue
        raise RuntimeError, "Could not find LLVM, try setting LIBCLANG to library!"
    end
end



# require 'ffi/clang' display stupid messages to stdout, silence them.
begin
    require "stringio"
    old_stdout, $stdout = $stdout, StringIO.new
    require 'ffi/clang'
    $stdout = old_stdout
end

# C-parsing module based on libclang and ffi/clang, i.e. ffi bindings.
#
# @example
#   require 'parsec'
#   ast = ParseC::SourceFile.new( './code_out.c', '.' )
#
module ParseC

    # Source file contains C-function definitions and header file
    # inclusions.
    class SourceFile

        # File name.
        attr_reader :name

        def SourceFile.parseLocalDefs( name, headerPath )
            file = SourceFile.new( name, headerPath )
            file.conf( :defsonly, true )
            file.conf( :mainonly, true )
            file.parse
            file
        end


        # Set file to parse and path for header search. Finally,
        # perform parsing for the file.
        def initialize( name, headerPath )
            @name = name
            @headerPath = headerPath

            @localHeader = {}
            @funcdefs = {}
            @funcalls = {}

            @conf = {}
            @conf[ :defsonly ] = false
            @conf[ :mainonly ] = false
        end


        def conf( opt, value = nil )
            if value
                @conf[ opt ] = value
            else
                @conf[ opt ]
            end
        end


        # Return list of included (local) header files. Can be used e.g. for
        # c-to-h file dependence.
        #
        # @return [Array<HeaderFile>] Headers.
        def headers
            @localHeader
        end


        # Return list of defined functions. Can be used e.g. to
        # generate function declarations to a header file
        # (automatically).
        #
        # @return [Array<HeaderFile>] Functions.
        def funcdefs
            @funcdefs
        end


        def funcalls
            @funcalls
        end


        # Return true if file name is for a local header.
        def isLocalHeader?( file )
            if file[0] != '/' && /\.h$/.match( file )
                true
            else
                false
            end
        end


        # Parse c-file.
        #
        # @param headerPath [String] Path for system header lookup.
        def parse( headerPath = @headerPath )

            index = FFI::Clang::Index.new
            translation_unit = index.parse_translation_unit( @name, headerPath )
            cursor = translation_unit.cursor

            # Get list of included local headers.
            translation_unit.inclusions do |header|
                if isLocalHeader? header
                    @localHeader[ header ] = HeaderFile.new( header )
                end
            end

            # Recursively travel the c-file hierarchy to detect
            # function definitions.
            cursor.visit_children do |cursor, parent|

                if cursor.kind == :cursor_function

                    # Check that function is in the file that is being
                    # analysed. Might also be from an included header.
                    if ( @conf[:mainonly] == false ) || cursor.location.from_main_file?

                        f = Function.new( cursor.spelling, cursor.variadic? )
                        f.setReturnType( cursor.result_type.spelling )

                        # Get argument list.
                        cursor.visit_children do |cursor, parent|

                            case cursor.kind

                            when :cursor_parm_decl
                                f.addArgument( cursor.spelling, cursor.type.spelling )

                            else
                                # puts cursor.kind
                                true

                            end

                            next :continue
                        end


                        # Mark function complete and add to list.
                        @funcdefs[ f.name ] = f.done

                        if @conf[:defsonly]
                            # Don't recurse anymore.
                            next :continue
                        else
                            next :recurse
                        end

                    else

                        next :continue

                    end

                elsif cursor.kind == :cursor_call_expr


                    # if cursor.location.from_main_file?
                    f = FunCall.new( cursor.spelling )
                    @funcalls[ f.name ] = f
                    # end

                    next :recurse

                else

                    # Non-functions are not interesting.
                    next :recurse

                end

            end
        end


        # Display c-file content.
        def display
            puts "File: #{@name}"
            puts "  Includes:"
            @localHeader.each_value do |i|
                puts "    #{i.name} | #{i.path}"
            end
            puts "  Functions:"
            @funcdefs.each_value do |f|
                puts "    #{f.declaration}"
            end
        end


        # Show raw content of c-file.
        def dump( headerPath = @headerPath )
            index = FFI::Clang::Index.new
            translation_unit = index.parse_translation_unit( @name, headerPath )
            cursor = translation_unit.cursor
            cursor.visit_children do |cursor, parent|
                puts "#{cursor.kind} #{cursor.spelling}"
                next :recurse
            end
        end


    end


    # Included local header.
    class HeaderFile

        # File name.
        attr_reader :name

        # File path.
        attr_reader :path

        # Initialize.
        def initialize( path )
            @name = File.basename( path )
            @path = path
        end

    end


    # Defined function.
    class Function

        # Func name.
        attr_reader :name

        # Variadic (end with ellipsis).
        attr_reader :variadic


        # Initialize function object. Default return type to void.
        #
        # @param name [String] Func name.
        def initialize( name, variadic = false )
            @name = name
            @variadic = variadic
            @args = {}
            @returnType = 'void'
        end


        # Set function return type.
        def setReturnType( type )
            @returnType = type
        end


        # Add argument to functions argument list.
        def addArgument( name, type )
            @args[ name ] = type
        end


        # Close function defition. If argument list is empty, then arg
        # list contains only void.
        def done
            if @variadic
                @args[ nil ] = '...'
            end

            if @args.empty?
                @args = { nil => 'void' }
            end
            self
        end


        # Get argument declarations String presentation.
        def argDecl( name )
            if name
                # Has type.
                "#{@args[ name ]} #{name}"
            else
                # Void/variadic.
                @args[ name ]
            end
        end


        # Return functions declation as String.
        def declaration
            "#{@returnType} #{@name}( #{(@args.keys.map { |k| argDecl(k) }).join(', ')} );"
        end

    end

    class FunCall
        # Func name.
        attr_reader :name

        def initialize( name )
            @name = name
        end
    end

end
